//---------------------------------------------------------
// RG SQL services
// Copyright © 2004 Ronald D. Grindle
//---------------------------------------------------------

package com.qanavigation.DBacc;

import com.qanavigation.Basics.*;

//import RG.MSSQL;


/**
* Implementation of an access to an SQL DB.
* <p>Native interface used.
* <p>DBAcc isA SQLRslt.
*/
public class DBacc2 extends SQLRslt
{

	/**
	* Last query commmand issued.
	*/
	public String m_LastQCmnd;
	
	protected IDBaccBase m_DBBase;
	
	 /**
	  * @param p_DBName
	  * @param p_Location
	  * @param p_User
	  * @param p_PW
	  */
	public DBacc2( String p_DBName, String p_Location, String p_User, String p_PW )
	{		
		setOK();
		
		m_DBBase = new MyDBacc( p_DBName, p_Location, p_User, p_PW );

		if( !checkRslt( m_DBBase ) ) {

			addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.DBacc(1)' failed." );
		}	
	}

	/**
	 * Insert data as a byte array to a (1) column of a table. The row is identified by 1 key column.
	 * @param p_KeyCol
	 * @param p_Key
	 * @param p_Content
	 * @return
	 */
	public SQLRslt addBLOB( String p_TblName, String p_KeyCol, String p_Key, String p_ContentCol, byte[] p_Content ) {
		
		SQLRslt res;
		
		res = m_DBBase.addBLOB( p_TblName, p_KeyCol, p_Key, p_ContentCol, p_Content );
   
		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.insertBLOB(..)' failed." );
		}
		
		return res;
	}
	

	/**
	 * 
	 * @return
	 */
	public SQLRslt close( ) {

		SQLRslt res;
		
		res = m_DBBase.close();

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.close(..)' failed." );
			
			setRsltStat( res );
		}
		
		return res;
	}
	
	/**
	 * 
	 * @return RsltStatVal.notFound if connection is lost.
	 */
	public SQLRslt chkConn() {
		
		return m_DBBase.chkConn();
	}
	
	/**
	 * Checks connection and tries to reconnect, if lost.
	 * @return
	 */
	public SQLRslt chkReconn() {
		
		return m_DBBase.chkReconn();
	}

	/**
	 * 
	 * @return
	 */
	public SQLRslt commit( ) {

		SQLRslt res;
		
		res = m_DBBase.commit();

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.commit(..)' failed." );
		}
		
		return res;
	}
	/**
	 * Decrements a counter. 
	 * @param p_TblName
	 * @param p_ColName
	 * @return
	 */
	public SQLRslt decrCounter( String p_TblName, String p_keyColName, String p_keyVal, String p_countColName ) {
		
		SQLRslt sqlRes;
		String query;
		
		query = "START TRANSACTION; ";
		
		sqlRes = execNoData( query );
		
		if( sqlRes.isOK() ) {
				
			query = "UPDATE " + p_TblName + " SET " + p_countColName + " = " + p_countColName + " - 1 ";
			query += " WHERE " + p_keyColName + " = '" + p_keyVal + "';";
			
			sqlRes = execNoData( query );
			
			if( sqlRes.isOK() ) {
				
				query = "COMMIT;";
				sqlRes = execNoData( query );
			}
			else {
				
				query = "ROLLBACK;";
				sqlRes = execNoData( query );
			}
		}
				
		return sqlRes;
	}
	
	/**
	 * Deletes a row.
	 * @param p_TableName
	 * @param p_WhereStmt
	 */
	public SQLRslt delete( String p_TableName, String p_WhereStmt ) {

		SQLRslt res;
		
		String query;
		
		query	= "DELETE FROM " + p_TableName + " WHERE " + p_WhereStmt;

		res = execNoData( query );

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.delete(..)' failed." );
		}
		
		return res;
	}
	
	/**
	 * Commits, if isOK, else rolls back.
	 * @param p_isOK
	 */
	public SQLRslt doCommit( boolean p_isOK ) {
		
		SQLRslt res;
		
		if( p_isOK ) {
			
			res = commit();
		}
		else {
			
			res = rollBack();
		}
		
		return res;
	}

	/**
	 * 
	 * @param p_Input
	 * @return the escaped string
	 */
	public String escapeStr2( String p_Input ) {
		
		String eStr;
		
		eStr = m_DBBase.escapeStr2( p_Input );
		
		return eStr;
	}

	//---------------------------------------------------------------------------
	protected SQLRslt execNoData( String p_Query ) {
		
		SQLRslt res;

		m_LastQCmnd = p_Query;

		res = m_DBBase.execNoData( p_Query );

		if( res.notOK() ) {

			res.addRsltAddInfo( "com.qanavigation.DBacc.DBacc.execNoData(..) for query '" + p_Query + "' failed." );
		}
		
		return res;
	}

	//---------------------------------------------------------------------------
	protected RsltTable execRead( String p_Query, String[] p_ColNames ) {

		RsltTable resT;

		m_LastQCmnd = p_Query;

		resT = m_DBBase.execRead( p_Query, p_ColNames );

		if( resT.notOK( true ) ) {

			resT.addRsltAddInfo( "com.qanavigation.DBacc.DBacc.execRead(..) for  query '" + p_Query + "' failed." );
		}
		
		return resT;
	}

	/**
	 * 
	 * @param p_WhereStmt
	 * @return
	 */
	public SetList findAllByVal( String p_TableName,  String p_ColName, String p_WhereStmt ) {
		
		SetList keyLst;
		
		String[] colNames = new String[1];
		colNames[0] = p_ColName;

		String query = "SELECT " + p_ColName + " FROM " + p_TableName;
		
		//select UIDpNode from NodeVal where Value LIKE '%a%';	
			
		query += " WHERE " + p_WhereStmt;

		keyLst = execRead( query, colNames ).toSetLst();

		if( keyLst.notOK( true ) ) {
			
			keyLst.addRsltAddInfo( m_LastQCmnd );

			keyLst.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.findAllByVal(..)' failed." );
		}

		return keyLst;
	}

	/**
	 * Free form select. Between p_HeadStmnt and p_TailStmnt the list of columns is inserted.
	 * @param p_HeadStmnt
	 * @param p_ColNames
	 * @param p_TailStmnt
	 * @return
	 */
	public RsltTable freeSelect( String p_HeadStmnt, String[] p_ColNames, String p_TailStmnt ) {

		int i, x;

		RsltTable resSel;

		String cols = "";
		String query = p_HeadStmnt;

		x = p_ColNames.length;

		for( i = 0; i < x; i++ ) {

			if( i > 0 ) {

				cols += ", ";
			}

			cols += p_ColNames[ i ];
		}
			
		query	+= cols;	
		query	+= p_TailStmnt;

		resSel = execRead( query, p_ColNames );

		if( resSel.notOK() ) {

			resSel.addRsltAddInfo( "com.qanavigation.DBacc.DBacc.freeSelect(..) for  query '" + p_HeadStmnt 
				+ "..." + p_TailStmnt + "'' failed. Cols: '" + cols + "'" );	
		}
		
		return resSel;
	}
	
	/**
	 * 
	 * @param p_TblName
	 * @param p_ColName
	 * @return
	 */
	public RsltString getIncrCounter( String p_TblName, String p_ColName ) {
		
		RsltString res;
		RsltTable resSel;
		
		SQLRslt sqlRes;
		String query;
		String[] colNames;
		
		res = new RsltString( true );
		
		colNames = new String[1];
		colNames[0] = p_ColName;
		
		query = "START TRANSACTION; ";
		
		sqlRes = execNoData( query );
		
		if( res.checkRslt( sqlRes )) {
			
			query = "SELECT " + p_ColName + " FROM " + p_TblName + " FOR UPDATE; ";
			resSel = execRead( query, colNames );

			if( res.checkRslt( resSel )) {
				
				query = "UPDATE " + p_TblName + " SET " + p_ColName + " = " + p_ColName + " + 1; ";
				sqlRes = execNoData( query );
			}
			
			if( res.checkRslt( sqlRes )) {
				
				query = "COMMIT;";
				sqlRes = execNoData( query );
			}
			else {
				
				query = "ROLLBACK;";
				sqlRes = execNoData( query );
			}
			
			if( res.checkRslt( sqlRes ) ) {
				
				res.setString( resSel.getCell( 0, 0 ) );
			}
		}
				
		return res;
	}

	/**
	 * 
	 * @param p_ColName
	 * @param p_TailStmnt
	 * @return
	 */
	public SetList getKeys( String p_ColName, String p_TailStmnt ) {

		SetList keyLst = new SetList();
		
		RsltTable rTbl;
		String[] colNames = new String[1];
		colNames[0] = p_ColName;

		String query = "SELECT ";

		query += p_ColName;
			
		query += p_TailStmnt;
		
		rTbl = execRead( query, colNames );
		
		keyLst.setRsltStat( rTbl );
		
		if( keyLst.isOK() ) {

			keyLst = rTbl.toSetLst();
		}

		if( keyLst.notOK( true ) ) {

			keyLst.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.getKeys(..)' failed." );
		}

		return keyLst;
	}
	
	/**
	 * Returns a single value identified by a single key.
	 * @param p_tblName
	 * @param p_valCol
	 * @param p_keyCol
	 * @param p_keyVal
	 * @return
	 */
	public RsltString getValByKey( String p_tblName, String p_valCol, String p_keyCol, String p_keyVal ) {
		
		RsltString res = new RsltString( true );
		
		String appx;
		String[] cols = new String[1];
		RsltTable rTbl;
		
		cols[0] = p_valCol;
		
		appx = "WHERE " + p_keyCol + " = '" + p_keyVal + "'";
		
		rTbl = select( p_tblName, cols, appx );
		
		if( res.checkRslt( rTbl ) ) {
			
			res.setString( rTbl.getCell( 0,0 ) );
		}

		if( res.notOK( true ) ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.getValByKey(..)' failed." );
		}
		
		return res;
	}
	
	/**
	 * Increments a counter.
	 * @param p_TblName
	 * @param p_ColName
	 * @return
	 */
	public SQLRslt incrCounter( String p_TblName, String p_keyColName, String p_keyVal, String p_countColName ) {
		
		SQLRslt sqlRes;
		String query;
		
		query = "START TRANSACTION; ";
		
		sqlRes = execNoData( query );
		
		if( sqlRes.isOK() ) {
				
			query = "UPDATE " + p_TblName + " SET " + p_countColName + " = " + p_countColName + " + 1 ";
			query += " WHERE " + p_keyColName + " = '" + p_keyVal + "';";
			
			sqlRes = execNoData( query );
			
			if( sqlRes.isOK() ) {
				
				query = "COMMIT;";
				sqlRes = execNoData( query );
			}
			else {
				
				query = "ROLLBACK;";
				sqlRes = execNoData( query );
			}
		}
				
		return sqlRes;
	}

	/**
	 * Insert a row.
	 * @param p_Qset
	 * @return
	 */
	public SQLRslt insert( QuerySet p_Qset ) {
		
		SQLRslt res;
		
		res = m_DBBase.insert( p_Qset );
   
		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.insert(..)' failed." );
		}

		return res;
	}
	
	/**
	* Inserts a new row or updates, if key is duplicate.
	* @param p_Qset
	* @return
	*/
	public SQLRslt insertUpdate( QuerySet p_Qset ) {

		SQLRslt res = new SQLRslt( true );
		
		res = insert( p_Qset );

		if( res.notOK() ) {

			if( res.getRsltStatVal() == RsltStatVal.isDupli ) { 
				
				res = update( p_Qset );
			}
		}
		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.insertUpdate(..)' failed." );
		}
		
		return res;
	}

	/**
	* Performs a roll back.
	* @return
	*/
	public SQLRslt rollBack( ) {

		SQLRslt res;
		
		res = m_DBBase.rollBack();

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.rollBack(..)' failed." );
		}
		
		return res;
	}

	/**
	 * Executes a standard select.
	 * @param p_TableName
	 * @param p_ColNames
	 * @param p_Appx
	 * @return
	 */
	public RsltTable select( String p_TableName, String[] p_ColNames, String p_Appx ) {

		RsltTable resSel;
		
		try {
		
			String selStmnt = "SELECT ";
				
			String appx = " FROM " + p_TableName + " " + p_Appx;
		
			resSel = freeSelect( selStmnt, p_ColNames, appx );
		
			if( resSel.notOK() ) {
		
				resSel.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.select(..)' failed." );
			}
		}
		catch( Exception excp ) {
			
			resSel = new RsltTable();
			
			resSel.setRsltByExcp( excp );
			resSel.addRsltAddInfo( "select() failed for '" + p_TableName + "', appx '" + p_Appx + "'" );
		}

		return resSel;
	}
	
	/**
	* Starts a transaction.
	* @return
	*/
	public SQLRslt startTrans( ) {

		SQLRslt res = m_DBBase.startTrans();
		
		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.startTrans(..)' failed." );
		}		
		
		return res;
	}	

	/**
	 * Updates a row. The 'WHERE' statement is automatically built 
	 * from the key columns from the QuerySet.
	 * @param p_Qset
	 * @return
	 */
	public SQLRslt update( QuerySet p_Qset ) {
		
		SQLRslt res;
		
		res = m_DBBase.update( p_Qset );
   
		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.DBacc.DBacc.update(..)' failed." );
		}

		return res;
	}
}
