//---------------------------------------------------------
//  RG Basic Tools Collection
//  Copyright © 2004 -2009 Ronald D. Grindle
//---------------------------------------------------------
/*
 * Created on Jun 28, 2004
 *
 */
package com.qanavigation.DBacc;

import java.sql.*;

import com.qanavigation.Basics.*;

/**
 * 
 *
 */
public class MyDBacc extends SQLRslt implements IDBaccBase
{

	public String ver = "V04.00.000 2009-07-27";
	
	protected String m_connStr;
	protected Connection m_Conn;
	protected Statement m_Stmt;
	
	/**
	 * Opens a MySQL database and initializes the service. 
	 * A URL is constructed, but the parameters are not escaped yet.
	 * @param p_DBName
	 * @param p_Location IP address host:port
	 * @param p_User may be "" (but not null);
	 * @param p_PW may be "" (but not null);
	 */
	public MyDBacc( String p_DBName, String p_Location, String p_User, String p_PW ) {
		
		try {
			
			setOK();
			
			Class.forName( "com.mysql.jdbc.Driver" );
			
			/*
			jdbc:mysql://[host:port],[host:port].../[database][?propertyName1][=propertyValue1][&propertyName2][=propertyValue2]...
			properties user, password
			*/
			//+ "?useServerPrepStmts=false" 
			
			m_connStr = "jdbc:mysql://" + p_Location + "/" + p_DBName; 
			
			if( p_User.length() > 0 ) {
				
				m_connStr += "?user=" + p_User;
			}
			
			if( p_PW.length() > 0 ) {
				
				m_connStr += "&password=" + p_PW;
			}

/*	
  Named pipes are a high performane alternative to TCP/IP on a Windows NT/2000/XP/2003 only environment
 	
			con = DriverManager.getConnection("jdbc:mysql:///test" +
			  "?socketFactory=com.mysql.jdbc.NamedPipeSocketFactory",
			  "root", "secret");		
*/			  	
			
			checkRslt( connect() );
		}
		catch( Exception excp ) {
			
			setRsltByExcp( excp );
		}
	}

	/**
	 * Add a byte stream
	 * @param p_TblName
	 * @param p_KeyCol
	 * @param p_Key
	 * @param p_ContentCol
	 * @param p_Content
	 * @return
	 */
	public SQLRslt addBLOB( String p_TblName, String p_KeyCol, String p_Key, String p_ContentCol, byte[] p_Content ) {

		SQLRslt res = new SQLRslt( true ); 
		
		try {
			
			String sqlStmnt;
			
			PreparedStatement prepStmnt;
			
			sqlStmnt = "update " + p_TblName + " set " + p_ContentCol + " = ? where " + p_KeyCol + " = '" + p_Key + "'";
			
			prepStmnt = m_Conn.prepareStatement(sqlStmnt);
	    
			prepStmnt.setBytes( 1, p_Content );
			
			prepStmnt.executeUpdate();
		}
		catch( SQLException sqlExcp ) {
		
			res.setRsltByExcp( sqlExcp );
			
			res.addRsltAddInfo( "'com.qanavigation.DBacc.MyDBacc.insertBLOB(..)' failed." );
		}
		
		return res;
	}
	
	/*
	 * @return RsltStatVal.notFound if connection is lost.
	 */
	public SQLRslt chkConn() {
		
		SQLRslt res = new SQLRslt( true );
		
		//boolean isOK;
		
		try {
			
			PreparedStatement prepStmnt;
			ResultSet resSet;
			
			prepStmnt = m_Conn.prepareStatement( "SELECT A FROM X" );
			
			resSet = prepStmnt.executeQuery();
			
			/* doesn't happen
			SQLWarning warning = resSet.getWarnings();
	  		
			if (warning != null) {
				
				res.setInError( "Check connection failed." );
	
				res.addRsltAddInfo("SQLWarning");
	
				while (warning != null) {
		
					res.addRsltAddInfo( "Message: " + warning.getMessage() );
					res.addRsltAddInfo( "SQLState: " + warning.getSQLState() );
					res.addRsltAddInfo( "Vendor error code: " + warning.getErrorCode() );
					warning = warning.getNextWarning();
				}
			}
			*/
		
			/* ----
			 * throws java.lang.AbstractMethodError
			
			isOK = m_Conn.isValid( 10 );
			
			if( !( isOK ) ) {
			
				res.setInError( "Connection not OK." );
			} */
		}
		catch( SQLException excp ) {
			
			boolean noConn, noX;
			int code;
			String state;
			
			state = excp.getSQLState();
			code = excp.getErrorCode();
			
			// Message "Communication link failure: java.io.EOFException, underlying cause: null"
			noConn = ( state.compareTo( "08S01" ) == 0 );
			
			if( noConn ) {
				
				res.setRsltStat( RsltStatVal.notFound );
			}
			else {
			
				// Message "Table 'eipwTst.X' doesn't exist" means connection is OK
				noX = (state.compareTo( "S1000" ) == 0) && (code == 1146);
				
				if( !( noX ) ) {
					
					res.addRsltAddFailInfo( "MyDBacc", "chkConn", "", "" );
					res.setRsltByExcp( excp );
				
					//res.addRsltAddInfo( "Last error code: " + res.getLastErrorCode() );
				}
			}
		}
		
		return res;
	}
	
	/**
	 * 
	 * @return Checks connection and tries to reconnect.
	 */
	public SQLRslt chkReconn( ) {
		
		SQLRslt res;
		
		res = chkConn();
		
		if( res.isInStat( RsltStatVal.notFound ) ) {
			
			close();
			
			res = connect();
		}
		
		return res;
	}
	
	// ------------------------------------
	/*
	private void addErrInf( IRslt p_Res ) {
		
		try {
			
			//SQLWarning sWrn = m_Conn.getWarnings();
			
			//p_Res.addRsltAddInfo( sWrn.getMessage() );
			
			p_Res.addRsltAddInfo( "Last Error Code: " + m_lastErrCode );
		}
		catch( Exception excp ) {

			p_Res.setRsltByExcp( excp );
		}
	}
	*/

	/**
	 * @return
	 */
	public SQLRslt close( ) {
	
		SQLRslt res = new SQLRslt( true );
		
		try {
			
			if( m_Stmt != null ) {
				
				m_Stmt.close();
			}
			if( m_Conn != null ) {
				
				m_Conn.close();
			}
		}
		catch( SQLException sqlExcp ) {
		
			setRsltByExcp( sqlExcp );
			res.setRsltByExcp( sqlExcp );
		}
		
		return res;
	}
	
	/**
	 * @return
	 */
	public SQLRslt commit( ) {

		SQLRslt res = new SQLRslt( true );
		
		try {
			
			m_Conn.commit();
		}
		catch( SQLException sqlExcp ) {
		
			res.setRsltByExcp( sqlExcp );
		}
		
		return res;
	}
	
	// -------------------------------------------
	private SQLRslt connect( ) {
		
		SQLRslt res = new SQLRslt( true );
		
		try {
			
			m_Conn = DriverManager.getConnection( m_connStr );
			
			if( m_Conn.isClosed() ) {
					
				res.setInError( "Connection to MySQL server for '" + m_connStr + "' failed ." );
			}
			else {
				
				m_Stmt = m_Conn.createStatement();
		
				m_Conn.setAutoCommit( false );
				
				//m_Stmt.setEscapeProcessing( true ); Should be on by default(?)			
			}
		}
		catch( SQLException sqlExcp ) {

			res.setRsltByExcp( sqlExcp );
		}
		catch( Exception excp ) {
			
			res.setRsltByExcp( excp );
		}
		
		return res;
	}
	
	/**
	 * Prepares a string for the DB.
	 * Escapes """, "'", "\".
	 * @param p_Str
	 * @return
	 */
	public String escapeStr2( String p_Str ) {
		
		String rStr;
		
		rStr = p_Str.replaceAll( "&", "&#038;"  ); //to be able to detect &#xx; vals later
		rStr = rStr.replaceAll( "\"", "&#034;"  );
		rStr = rStr.replaceAll( "'", "&#039;"  );
		
		rStr = rStr.replaceAll( "~", "&#126;"  ); // work around (see unescape) 	
		rStr = rStr.replaceAll( "\\\\", "~"  ); //&#092; causes trouble (see unescape)
		
		//droped rStr = "'" + rStr + "'";
		
		return rStr;					
	}
	
	/**
	 * For queries that do not add data or expect any result back. 
	 * @param p_Query
	 * @return
	 */
	public SQLRslt execNoData( String p_Query ) {
		
		SQLRslt res = new SQLRslt( true );
		
		try {
			
			m_Stmt.execute( p_Query );
			
			// never seems to happen...
			SQLWarning warning = m_Stmt.getWarnings();
			
			if (warning != null) {
				
				res.addRsltAddInfo("SQLWarning");
				
				while (warning != null) {
					
					res.addRsltAddInfo( "Message: " + warning.getMessage() );
					res.addRsltAddInfo( "SQLState: " + warning.getSQLState() );
					res.addRsltAddInfo( "Vendor error code: " + warning.getErrorCode() );
					warning = warning.getNextWarning();
				}
			}
		}
		catch( SQLException sqlExcp ) {
			
			res.setRsltByExcp( sqlExcp );
			//addErrInf( res );
		}
	
		return res;
	}

	/**
	 * forced to public by interface
	 */
	public RsltTable execRead( String p_Query, String[] p_ColNames ) {
			
		RsltTable resT = new RsltTable();
		
		try {
			
			//m_Stmt = m_Conn.createStatement();
			
			ResultSet resSet;
			
			resSet = m_Stmt.executeQuery( p_Query );
		
			SQLWarning warning = resSet.getWarnings();
  		
			if (warning != null) {
	
				resT.addRsltAddInfo("SQLWarning");
	
				while (warning != null) {
		
					resT.addRsltAddInfo( "Message: " + warning.getMessage() );
					resT.addRsltAddInfo( "SQLState: " + warning.getSQLState() );
					resT.addRsltAddInfo( "Vendor error code: " + warning.getErrorCode() );
					warning = warning.getNextWarning();
				}
			}		
			
			setRTbl( resT, resSet, p_ColNames.length );
		}
		catch( SQLException sqlExcp ) {
			
			SQLRslt sRes = new SQLRslt();
		
			sRes.setRsltByExcp( sqlExcp );
			
			resT.setRsltStat( sRes, "MyDBacc.execRead() failed." );
		}		
		
		return resT;
	}
	
	
	// ----------------------------------------
	protected void finalize( ) throws Throwable {
		
		m_Stmt.close();
		m_Conn.close();
		
		super.finalize();
	}
		
	/**
	 * @param p_Qset
	 * @return
	 */
	public SQLRslt insert( QuerySet p_Qset ) {
		
		SQLRslt res = new SQLRslt( true ); 
		
		byte[] bVal;
		     
		int i, k, rows;
		
		PreparedStatement prepStmnt;
		
		String sqlStmnt = "";

		try {
			
			sqlStmnt = "INSERT INTO " +  p_Qset.m_TableName + " (";

			k = 0;

			rows = p_Qset.getRowNum();
			
			for( i = 0; i < rows; ++i ) {

				if( i > 0 ) {

					sqlStmnt += ", ";
				}
				
				sqlStmnt += p_Qset.getQColName( i );
			}

			sqlStmnt += ") VALUES ( ";	
			
			for( i = 0; i < rows; ++i ) {

				if( i > 0 ) {

					sqlStmnt += ", ";
				}
				
				sqlStmnt += "?";
			}
			
			sqlStmnt += " )";

			prepStmnt = m_Conn.prepareStatement( sqlStmnt );

			k = 1;

			for( i = 0; i < rows; ++i ) {

				if( p_Qset.isBlob( i ) ) {

					bVal = escapeStr2( p_Qset.getQColVal( i ) ).getBytes();

					prepStmnt.setBytes( k, bVal );
				}
				else {

					prepStmnt.setString( k, escapeStr2( p_Qset.getQColVal( i )) );
				}

				k++;
			}

			prepStmnt.executeUpdate();
		}
		catch( SQLException sqlExcp ) {

			res.setRsltByExcp( sqlExcp );
		}
		
		if( res.notOK() ) {
			
			if( lastInsertDbl( res ) ) {
				
				res.setRsltStat( RsltStatVal.isDupli );
			}
			else {
				
				res.addRsltAddInfo( "SQL: '" + sqlStmnt + "'." );
				
				rows = p_Qset.getRowNum();

				for( i = 0; i < rows; ++i ) {
					
					res.addRsltAddInfo( escapeStr2( p_Qset.getQColVal( i ) ) );
				}
				
				res.addRsltAddInfo( "'com.qanavigation.DBacc.MyDBacc.insert(..)' failed." );
			}
		}
		
		return res;		
	}
	
	/**
	 * @param p_SQLRes
	 * @return
	 */
	public boolean lastInsertDbl( ISQLRslt p_SQLRes ) {

		boolean isDbl = false;

		if( 1062 == p_SQLRes.getLastErrorCode() ) {

			isDbl = true;
		}

		return isDbl;
	}
	
	/**
	 * 
	 */
	public void resetRslt() {
		
		try {
			
			m_Conn.clearWarnings();
			m_Stmt.clearWarnings();
		}
		catch( SQLException sqlExcp ) {
		
			setRsltByExcp( sqlExcp );
		}
				
		super.resetRslt();
	}
	
	/**
	 * @return
	 */
	public SQLRslt rollBack( ) {

		SQLRslt res = new SQLRslt( true );
		
		try {
			
			m_Conn.rollback();
		}
		catch( SQLException sqlExcp ) {
		
			res.setRsltByExcp( sqlExcp );
		}
		
		return res;
	}

	//-----------------
	// Converts any null pointer result into an empty line!
	// unescapes values.
	private void setRTbl( RsltTable p_RTbl, ResultSet p_ResSet, int p_ColNum ) {
		
		boolean ok;
		int i, x;
		int rowSize;
		String val;
		String[] cols;
		
		p_RTbl.setColSize( p_ColNum );
		
		rowSize = 0;
		x = p_ColNum;
		
		try {
					
			//int numberOfRows = p_ResSet.getRow();	
						
			ok = p_ResSet.first();
			
			while( ok ) {
					
				cols = new String[x];
								
				for( i = 0; i < x; ++i ) {
					
					// Index starts at 1 !!!!!!!!!!!!!!!!
					val = p_ResSet.getString( i + 1 );
					
					if( val == null ) {
					
						cols[i] = "";
					}
					else {
						
						cols[i] = unescapeStr( val );
					}
				}
				
				p_RTbl.add( cols );
				
				rowSize++;
				
				ok = p_ResSet.next();
			}
	
			if( rowSize < 1 ) {
	
				p_RTbl.setRsltStat( RsltStatVal.isEmpty );
			}
			else {
	
				p_RTbl.setRowSize( rowSize );
				
				p_RTbl.setOK();	
			}		
		}
		catch( SQLException excp ) {
			
			p_RTbl.setRsltByExcp( excp ); 
		}
	}
	

	/**
	 * Not required for MySQL
	 * @return
	 */
	public SQLRslt startTrans() {

		SQLRslt res = new SQLRslt( true );
		
		return res;
	}
	
	/**
	 * @param p_Str
	 * @return
	 */
	public String unescapeStr( String p_Str ) {
		
		String rStr;
		
		//crashes: rStr = p_Str.replaceAll( "&#092;", "\\"  );
		//workaround:
		rStr = p_Str.replaceAll( "~", "\\\\"  );	
		rStr = rStr.replaceAll( "&#126;", "~"  );
				
		rStr = rStr.replaceAll( "&#039;", "'"  ); 
		rStr = rStr.replaceAll( "&#034;", "\""  );	
		
		rStr = rStr.replaceAll( "&#038;", "&" );
	
		return rStr;					
	}	
	
	/**
	 * @param p_Qset
	 * @return
	 */
	public SQLRslt update( QuerySet p_Qset ) {

		SQLRslt res = new SQLRslt( true ); 
		
		byte[] bVal;
		     
		int i, k, rows;
		
		PreparedStatement prepStmnt;
		
		String sqlStmnt = "";

		try {

			sqlStmnt = "UPDATE " + p_Qset.m_TableName + " SET ";

			k = 0;

			rows = p_Qset.getRowNum();

			for( i = 0; i < rows; ++i ) {
				
				if( !(p_Qset.isKey( i )) ) {

					if( k > 0 ) {

						sqlStmnt += ", ";
					}

					sqlStmnt += p_Qset.getQColName( i );
					sqlStmnt += " = ?";

					k++;
				}
			}

			k = 0;

			for( i = 0; i < rows; ++i ) {

				if( p_Qset.isKey( i ) ) {

					if( k < 1 ) {

						sqlStmnt += " WHERE ";
					}

					if( k > 0 ) {

						sqlStmnt += " AND ";
					}

					sqlStmnt += p_Qset.getQColName( i );
					sqlStmnt += " = '" + escapeStr2( p_Qset.getQColVal( i ) ) + "'";

					k++;
				}
			}

			prepStmnt = m_Conn.prepareStatement( sqlStmnt );

			k = 1;

			for( i = 0; i < rows; ++i ) {
				
				if( !(p_Qset.isKey( i )) ) {
					
					if( p_Qset.isBlob( i ) ) {

						bVal = escapeStr2( p_Qset.getQColVal( i ) ).getBytes();

						prepStmnt.setBytes( k, bVal );
					}
					else {

						prepStmnt.setString( k, escapeStr2( p_Qset.getQColVal( i )) );
					}

					k++;
				}
			}

			prepStmnt.executeUpdate();
		}
		catch( SQLException sqlExcp ) {

			res.setRsltByExcp( sqlExcp );
			
			res.addRsltAddInfo( "SQL: '" + sqlStmnt + "'." );
			
			rows = p_Qset.getRowNum();
			for( i = 0; i < rows; ++i ) {
				
				res.addRsltAddInfo( escapeStr2( p_Qset.getQColVal( i ) ) );
			}			

			res.addRsltAddInfo( "'com.qanavigation.DBacc.MyDBacc.update(..)' failed." );
		}
		
		return res;
	}	
}
