//---------------------------------------------------------
// RG XML DB Tools Collection
// Copyright © 2003 - 2009 Ronald D. Grindle
//---------------------------------------------------------

package com.qanavigation.XMLDB;

import java.io.File;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.qanavigation.Basics.Rslt;
import com.qanavigation.Basics.Rslt2DArray;
import com.qanavigation.Basics.RsltStatVal;
import com.qanavigation.Basics.RsltString;
import com.qanavigation.Basics.RsltTable;
import com.qanavigation.Basics.SetList;
import com.qanavigation.DBacc.DBacc2;
import com.qanavigation.DBacc.QuerySet;
import com.qanavigation.UIDserv.UIDclient;

/**
 * Class for handling XMLs in a DB.
 *	<remarks>
 * The concept of an XML file has been altered for this implementation.
 * The topmost XML element here is also the parent element for all other elements, 
 * such as the document information. This has been done for the sake 
 * of the DB design. This allows any element (node, attribute, document info etc.) to have
 * only one reference to a parent node and still enables the elements to be linked consistently 
 * without any ambiguity and still providing full navigability.
 * 
 * 'get..' service calls do not spoil the status of this object on failure. 
 * The actual status is found with the object returned.
 * </remarks>
 * XMLDBbasics isA Rslt
 */
public class XMLDBbasics extends Rslt
{

	public DBacc2 m_DB;
	protected UIDclient m_UIDServ;

	// --------------------------------------------------------------------------
	/* To solve the "too many connections" problem, only import the DBacc objekt is allowed.
	protected XMLDBbasics( String p_DBName, String p_Location, String p_User, String p_PW, String p_UIDservHostName  ) {

		setOK();
		
		m_DB = new DBacc( p_DBName, p_Location, p_User, p_PW );
		
		if( checkRslt( m_DB ) ) {

			m_UIDServ = new UIDclient( p_UIDservHostName );
		}
		else {

			addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.XMLDBbasics([1])' failed." );
		}		
	}
	*/
	
	// --------------------------------------------------------------------------
	protected XMLDBbasics( DBacc2 p_DBAcc, String p_UIDservHostName, int p_port  ) {

		setOK();
		
		m_DB = p_DBAcc;
		
		if( checkRslt( m_DB ) ) {

			m_UIDServ = new UIDclient( p_UIDservHostName, p_port );
		}
		else {

			addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.XMLDBbasics([1])' failed." );
		}		
	}	

	// --------------------------------------------------------------------------
	// Copy constructor
	protected XMLDBbasics( XMLDBbasics p_xDB ) {

		setOK();
		
		if( checkRslt( p_xDB ) ) {
			
			m_DB = p_xDB.m_DB;
			
			m_UIDServ = p_xDB.m_UIDServ;
		}
		else {

			addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.XMLDBbasics([2])' failed." );
		}			
	}
	
	/*
	 * Appends a file to a Node. Stores only the information about the link. The actual file 
	 * has to be stored somewhere else. 
	 * @param p_parentUID
	 * @param p_DateTime
	 * @param p_Path
	 * @param p_Name
	 * @param p_Content
	 * @return The UID of the link.
	 *
	public RsltString appendFile( String p_parentUID, String p_DateTime, String p_Path, String p_Name ) {
		
		/* FileUpLoad(	UID CHAR(27) UNIQUE NOT NULL,
										UIDpNode CHAR(27) NOT NULL,
										DateTime CHAR(10),
										Path BLOB,
										Name VARCHAR(255) NOT NULL,
										Content BLOB,
		 * /
		
		QuerySet qSet;
		RsltString UID;
		
		UID = m_UIDServ.getUID();
		
		if( UID.isOK() ) {
			
			startTrans();
			
			qSet = new QuerySet( 5, XMLDBDefs.c_TblNmFileUp );
			
			qSet.addQCol( XMLDBDefs.c_ColNmParentUID, p_parentUID );
	
			qSet.addQCol( XMLDBDefs.c_ColNmUID, UID.getString() );
			
			qSet.addQCol( XMLDBDefs.c_ColNmDateTime, p_DateTime );
			
			qSet.addQCol( XMLDBDefs.c_ColNmPath, p_Path );
			
			qSet.addQCol( XMLDBDefs.c_ColNmName, p_Name );
	
			m_DB.insert( qSet );
		}

		/* stored as file
		if( UID.checkRslt( m_DB ) ) {
			
			m_DB.addBLOB( XMLDBDefs.c_TblNmFileUp, XMLDBDefs.c_ColNmUID, UID.getString(), XMLDBDefs.c_ColNmContent, p_Content );
		}
		* /
		
		if( !UID.checkRslt( m_DB ) ) {

			UID.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeValNode2DB(..)' failed." );
			
			rollBack();
		}
		else {
			
			commit();
		}
		
		return UID;
	}*/
	
	/**
	 * 
	 * @param p_UID
	 * @return
	 */
	public Rslt chkDoc( String p_UID ) {
		
		Rslt res = new Rslt( true );
		
		RsltTable valTbl;

		String tblName = XMLDBDefs.c_TblNmNode;

		String[] cols = { XMLDBDefs.c_ColNmUID };

		String appx = "WHERE " + XMLDBDefs.c_ColNmUID + " = '" + p_UID + "'";

		valTbl = m_DB.select( tblName, cols, appx );
		
		if( !( res.checkRslt( valTbl ) ) ) {
			
			res.addRsltAddInfo( "chkDoc(): Node for UID '" + p_UID + "' could not be found." );
		}
		
		return res;
	}

	/**
	 * The DB action is commited.
	 */
	protected Rslt commit( ) { 

		Rslt res;
		
		res = m_DB.commit();

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.commit()' failed." );
		}
		
		return res;
	}

	// --------------------------------------------------------------------------
	protected Rslt delAttrs( String p_UID ){

		Rslt res;
		
		String whereClause;
		String tblName = XMLDBDefs.c_TblNmAttrib;

		whereClause = XMLDBDefs.c_ColNmParentUID + " = '" + p_UID + "'";

		res = m_DB.delete( tblName, whereClause );
		
		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.delAttrs(..)' failed." );
		}
		
		return res;
	}

	//---------------------------------------------------------------------------
	/**
	 * Deletes a node.
	 * Runs through the list of subnodes. Normally not required for MySQL.
	 */
	public Rslt deleteNode( String p_UID ) {

		int i, x;
		
		Rslt res = new Rslt( true );
		Rslt aRes;
		
		SetList subUIDLst;
		String whereClause;
		String tblName = XMLDBDefs.c_TblNmNode;

		subUIDLst = getUIDLstByParent( "", p_UID );

		if( res.checkRslt( subUIDLst, true ) ) {

			x = subUIDLst.size();

			if( x > 0 ) {

				String UID;

				for( i = 0; i < x; ++i ) {

					UID = subUIDLst.getString(i);
	
					aRes = deleteNode( UID );

					if( !res.checkRslt( aRes ) ) {

						i = x;
					}
				}
			}
			
			whereClause = XMLDBDefs.c_ColNmUID + " = '" + p_UID + "'";

			aRes = m_DB.delete( tblName, whereClause );
			
			res.checkRslt( aRes );
		}

		if( res.notOK() ) {

			addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.deleteNode(..)' failed." );
		}
		
		return res;
	}

	/**
	 * Removes an entire XML from the DB. 
	 */
	public Rslt deleteXML( String p_UID ){

		startTrans();
		
		Rslt res;
		
		res = deleteNode( p_UID );

		if( res.notOK() ) {

			addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.deleteXML(..)' failed." );
			
			rollBack();
		}
		else {
			
			commit();
		}
		
		return res;
	}

	
	/** 
	 * @param p_UID
	 * @param p_AttrName
	 * @return The attributes value
	 */
	public RsltString getAttrVal( String p_UID, String p_AttrName ) { 

		int x;

		RsltString attrVal;

		RsltTable attrTbl;

		attrVal = new RsltString( true );

		String tblName = XMLDBDefs.c_TblNmAttrib;

		String[] cols = { XMLDBDefs.c_ColNmValue };

		String appx = "WHERE " + XMLDBDefs.c_ColNmParentUID + " = '" + p_UID + "'"
					+ " AND " + XMLDBDefs.c_ColNmName + " = '" + p_AttrName + "'";

		attrTbl = m_DB.select( tblName, cols, appx );

		if( attrVal.checkRslt( attrTbl ) ) {
			
			x = attrTbl.getRowNum();
			
			if( x > 1 ) {
				
				attrVal.setInError( "Too many matching attributes found for '" + p_AttrName + "'." );
			}
			else {
				
				if( x == 0 ) {
					
					attrVal.setRsltStat( RsltStatVal.isEmpty, "" );
				}
				else {
					
					attrVal.setString(attrTbl.getCell( 0, 0 ) );
				}
			}
		}

		if( attrVal.notOK() ) {

			attrVal.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getAttrVal(..)' failed." );
		}

		return attrVal;
	}
	
	// ---------------------
	// by default compared case insensitive
	private String getCmpStmnt( String p_valName, boolean p_withCmp, Rslt2DArray p_AttrSet, int p_i ) {
		
		String cmpStmt;
		String op;
		String cmpIsNum;
		
		String cVal, sVal;
		
		cVal = "LOWER( " + p_valName + " )";
		sVal =  p_AttrSet.getCell( 1, p_i ).toLowerCase();
		
		if( p_withCmp ) {
			
			op = p_AttrSet.getCell( 2, p_i );			
			
			if( op.compareTo( "LIKE" ) == 0 ) {
				
				cmpStmt = cVal + " LIKE '%" + m_DB.escapeStr2( sVal ) + "%'" ;
			}
			else if( ( op.compareTo( "<=" ) == 0 ) || ( op.compareTo( ">=" ) == 0 ) ) {
				
				cmpIsNum = p_AttrSet.getCell( 3, p_i );
				
				if( cmpIsNum.compareTo( "true" ) == 0 ) {
					
					cmpStmt = "1*(" + cVal + ") " + op + " 1*(" + m_DB.escapeStr2( sVal ) + ")";					
				}
				else {
			
					cmpStmt = cVal + " " + op + " '" + m_DB.escapeStr2( sVal ) + "'";
				}
			}
			else {
				
				cmpStmt = cVal + " = '" + m_DB.escapeStr2( sVal ) + "'";
			}
		}
		else {
			
			cmpStmt = cVal + " = '" +  m_DB.escapeStr2( sVal ) + "'";
		}
		
		if( true ) {		
			
			cmpStmt += " COLLATE latin1_general_ci";
		}
		
		return cmpStmt;
	}
	
	/**
	 * 
	 * @param p_parentUID
	 * @return String[0]: dtdName, String[1]: pubDTDID,
	 */
	public String[] getDTDInfo( String p_parentUID ) { 
		
		int pos, nPos;
		
		String[] dTDInfo = { "", "" };
		
		try {
			
			RsltTable rT = getNodeLstByElemType( p_parentUID, Node.DOCUMENT_TYPE_NODE );
			
			if( rT.isNotEmpty() ) {

				String UID = rT.getCell( 0, 0 );
			
				RsltString val = getNodeVal( UID );

				pos = val.getString().indexOf( "SYSTEM" );

				if( pos > -1 ) {

					nPos = pos + "SYSTEM".length() + 1;

					dTDInfo[0] = val.getString().substring( nPos );
				}

				pos = val.getString().indexOf( "PUBLIC" );
	
				if( pos > -1 ) {
	
					nPos = pos + "PUBLIC".length() + 1;
	
					dTDInfo[1] = val.getString().substring( nPos );
				}
			}
		}
		catch( Exception excp ) {

			setRsltByExcp( excp );
		}

		return dTDInfo;
	}

	/**
	 * 
	 * @param p_UID
	 * @return RsltTable; Cols: Name, Path.
	 */
	public RsltTable getFileInfo( String p_UID ) {

		RsltTable fInfoTbl;

		String tblName = XMLDBDefs.c_TblNmXMLFile;

		String[] cols = { XMLDBDefs.c_ColNmName, XMLDBDefs.c_ColNmPath };

		String appx = "WHERE " + XMLDBDefs.c_ColNmParentUID + " = '" + p_UID + "'";

		fInfoTbl = m_DB.select( tblName, cols, appx );
		
		if( fInfoTbl.notOK() ) {

			fInfoTbl.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getFileInfo(..)' failed." );
		}

		return fInfoTbl;
	}

	/**
	 * Returns the last query command issued. For debugging purposes.
	 * @return
	 */
	public String getLastQCommand( ) {

		return m_DB.m_LastQCmnd;
	}

	/**
	 * Should not deliver more than one! (Not restricted yet.)
	 * @param p_UID
	 * @return RsltTable; Col sequence: pUID, Name, Type.
	 */
	public RsltTable getNodeAsTable( String p_UID ) {

		RsltTable nodeTbl;

		String tblName = XMLDBDefs.c_TblNmNode;

		String[] cols = { XMLDBDefs.c_ColNmParentUID, XMLDBDefs.c_ColNmName, XMLDBDefs.c_ColNmNodeType };

		String appx = "WHERE " + XMLDBDefs.c_ColNmUID + " = '" + p_UID + "'";

		nodeTbl = m_DB.select( tblName, cols, appx );

		if( nodeTbl.notOK() ) {

			nodeTbl.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodeAsTable(..)' failed." );
		}

		return nodeTbl;
	}
	
	/**
	 * Gets the list of attributes for a node as table.
	 * @param p_UID
	 * @return RsltTable; Cols: Name, Value.
	 */
	public RsltTable getNodeAttrs( String p_UID ) { 

		RsltTable attrTbl;

		String tblName = XMLDBDefs.c_TblNmAttrib;

		String[] cols = { XMLDBDefs.c_ColNmName, XMLDBDefs.c_ColNmValue };

		String appx = "WHERE " + XMLDBDefs.c_ColNmParentUID + " = '" + p_UID + "'";

		attrTbl = m_DB.select( tblName, cols, appx );

		if( attrTbl.notOK() ) {

			attrTbl.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodeAttrs(..)' failed." );
		}

		return attrTbl;
	}

	/**
	 * 
	 * @param p_ParUID
	 * @param p_eType
	 * @return RsltTable; Cols: UID, Name, Type.
	 */
	protected RsltTable getNodeLstByElemType( String p_ParUID, short p_eType ) {
	
		RsltTable nodeTbl;
	
		String tblName = XMLDBDefs.c_TblNmNode;
	
		String[] cols = { XMLDBDefs.c_ColNmUID, XMLDBDefs.c_ColNmName, XMLDBDefs.c_ColNmNodeType };
	
		String appx = "WHERE " + XMLDBDefs.c_ColNmParentUID + " = '" + p_ParUID + "'"
							+ " AND " + XMLDBDefs.c_ColNmNodeType + " = '" + String.valueOf(p_eType) + "'";
	
		nodeTbl = m_DB.select( tblName, cols, appx );
	
		if( nodeTbl.notOK() ) {
	
			nodeTbl.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getXNodeLstByParent(..)' failed." );
		}
	
		return nodeTbl;
	}

	// --------------------------------------------------------------------------
	/* somthing better in Java?
	protected RsltString escapeString( String p_Unescaped ) {

		RsltString resStr = new RsltString();

		String escapedStr;

		escapedStr = p_Unescaped.Replace( @"\", @"\\" );
		escapedStr = escapedStr.Replace( "'", @"\'" );
		//goes bang escapedStr = escapedStr.Replace( "\"", @"\"" );
		//escapedStr = escapedStr.Replace( '\n', @'\n' );
		//escapedStr = escapedStr.Replace( '\r', @'\r'" );
		escapedStr = escapedStr.Replace( "%", @"\%" );
		escapedStr = escapedStr.Replace( "\t", @"\t" );
		escapedStr = escapedStr.Replace( "_", @"\_" );

		resStr.setString( escapedStr;

		return resStr;
	}
	*/

	/**
	 * @param p_UID
	 */
	public RsltString getNodeName( String p_UID ) {

		RsltTable valTbl;
		RsltString val = new RsltString( true );

		String tblName = XMLDBDefs.c_TblNmNode;

		String[] cols = { XMLDBDefs.c_ColNmName };

		String appx = "WHERE " + XMLDBDefs.c_ColNmUID + " = '" + p_UID + "'";

		valTbl = m_DB.select( tblName, cols, appx );

		if( val.checkRslt( valTbl ) ) {

			val.setString( valTbl.getCell( 0, 0 ) );
		}

		if( val.notOK() ) {

			val.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodeName()' failed." );
		}

		return val;
	}

	/**
	 * Retrieves only elements
	 * @param p_ElemType
	 * @return SetList of UIDs.
	 */
	public SetList getNodes( String p_ElemType ) {
	 
		SetList UIDLst;

		/* SELECT Node.UID
			 FROM Node
			 WHERE Node.Name = 'TestCase'
		*/

		//int iElem = (int)Node.Element;
		short iElem = Node.ELEMENT_NODE;
		
		String col = XMLDBDefs.c_ColNmUID;

		String qTail = " FROM " + XMLDBDefs.c_TblNmNode;

		/* WHERE Node.Name = 'TestCase' */

		qTail += " WHERE " + XMLDBDefs.c_ColNmName + " = '" + p_ElemType + "'";
		qTail += " AND " + XMLDBDefs.c_ColNmNodeType + " = '" + String.valueOf( iElem ) + "'";

		UIDLst = m_DB.getKeys( col, qTail );
		
		if( UIDLst.notOK( true ) ) {

			UIDLst.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodes(..)' failed." );
		}

		return UIDLst;
	}

/*2004-05-11 causes trouble with MSSQL
	public SetList getNodesByAttr( String p_ElemType, Rslt2DArray p_AttrSet ) {
	 

		int i, x;
		SetList UIDLst;
		String[] vTabNames;
		String prep;
		char chr;

		x = p_AttrSet.getRowNum();

		if( x > 26 ) {
// throws!
			throw( new ApplicationException( "Not prepared for attributes more than 26." ) );
		}

		vTabNames = new String[ x ];

		for( i = 0; i < x; ++i ) {

			chr = (char)(i + 65);
			vTabNames[i] = chr.ToString()+ "tab";
		}

		/* SELECT Node.UID
			 FROM Node
			 LEFT JOIN  Attrib AS A, Attrib AS B, Attrib AS C 
			 ON Node.UID=A.UIDpNode
			 WHERE Node.Name = 'TestCase'
			 AND A.UIDpNode = B.UIDpNode AND A.UIDpNode = C.UIDpNode
			 AND (A.Name = 'TCType' AND A.Value = 'mo')
			 AND (B.Name = 'Author' AND B.Value = 'sdg')
			 AND (C.Name = 'Time' AND C.Value = '17:52'); *//*

		String qHead = "SELECT ";
		
		String col = XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmUID;

		String qTail = " FROM " + XMLDBDefs.c_TblNmNode;

		//LEFT JOIN  Attrib AS A, Attrib AS B, Attrib AS C 

		qTail += " LEFT JOIN";

		prep = " " + XMLDBDefs.c_TblNmAttrib + " AS ";

		for( i = 0; i < x; ++i ) {

			if( i > 0 ) {

				qTail += ",";
			}

			qTail += prep + vTabNames[i];
		}

		//ON Node.UID=A.UIDpNode
		qTail += " ON " + XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmUID
			+ " = " + vTabNames[0] + "." + XMLDBDefs.c_ColNmParentUID;

		/* WHERE Node.Name = 'TestCase'
			 AND A.UIDpNode = B.UIDpNode AND A.UIDpNode = C.UIDpNode
			 AND (A.Name = 'TCType' AND A.Value = 'mo')
			 AND (B.Name = 'Author' AND B.Value = 'sdg')
			 AND (C.Name = 'Time' AND C.Value = '17:52'); *//*

		qTail += " WHERE ";

		qTail += XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmName + " = '" + p_ElemType + "'";

		// AND A.UIDpNode = B.UIDpNode AND A.UIDpNode = C.UIDpNode

		for( i = 1; i < x; ++i ) {

			qTail += " AND " + vTabNames[0] + "." + XMLDBDefs.c_ColNmParentUID + " = " 
				+ vTabNames[i] + "." + XMLDBDefs.c_ColNmParentUID;
		}

		/* AND (A.Name = 'TCType' AND A.Value = 'mo')
			 AND (B.Name = 'Author' AND B.Value = 'sdg')
			 AND (C.Name = 'Time' AND C.Value = '17:52') *//*

		for( i = 0; i < x; ++i ) {
			
			qTail += " AND (" + vTabNames[i] + "." + XMLDBDefs.c_ColNmName + " = '" +
				p_AttrSet.getCell( 0, i ) + "' AND "
				+ vTabNames[i] + "." + XMLDBDefs.c_ColNmValue + " = '"
				+ p_AttrSet.getCell( 1, i ) + "')"; 
		}

		UIDLst = m_DB.getKeys( qHead, col, qTail );

		return UIDLst;
	}			
*/
	
	/**
	 * Provide the attributes and the values searched for in the QuerySet.
	 * @param p_AttrSet
	 * @return SetList of UIDs.
	 */
	public SetList getNodesByAttr( Rslt2DArray p_AttrSet ) {
	
		boolean withCmp;
		
		int i, x;
		SetList UIDLst;
		SetList UIDLst2;
		
		String cmp;
		String qTail;

		withCmp = p_AttrSet.getColSize() > 2;
		
		x = p_AttrSet.getRowNum();

		/* SELECT UIDpNode
			 FROM Attrib 
			 WHERE Name = 'Prio' AND Value = '1' */

		String qMid = " FROM " + XMLDBDefs.c_TblNmAttrib;

		if( x > 0 ) {
			
			cmp = getCmpStmnt( XMLDBDefs.c_ColNmValue, withCmp, p_AttrSet, 0 );
			
			// AND (Attrib.Name = 'Prio' AND Attrib.Value = '1')
			qTail = qMid 
				+ " WHERE ( " + XMLDBDefs.c_ColNmName + " = '" + p_AttrSet.getCell( 0, 0 ) + "'" 
				+ " AND " + cmp + " )";
		}
		else {
			
			qTail = qMid;		
		}

		UIDLst = m_DB.getKeys( XMLDBDefs.c_ColNmParentUID, qTail );
//UIDLst.addRsltAddInfo( qTail );	

		for( i = 1; (i < x) && UIDLst.isOK( true ); ++i ) {
			
			cmp = getCmpStmnt( XMLDBDefs.c_ColNmValue, withCmp, p_AttrSet, i );

			// AND (Name = 'Prio' AND Value = '1')
			qTail = qMid 
				+ " WHERE ( " + XMLDBDefs.c_ColNmName + " = '" + p_AttrSet.getCell( 0, i ) + "'"
				+ " AND " + cmp + " )";

			UIDLst2 = m_DB.getKeys( XMLDBDefs.c_ColNmParentUID, qTail );

			if( UIDLst.checkRslt( UIDLst2, true ) ) {

				UIDLst.intersect( UIDLst2 );
			}
//UIDLst.addRsltAddInfo( qTail );
		}

		if( UIDLst.notOK() ) {

			UIDLst.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodesByAttr(..)' failed." );
		}

		return UIDLst;
	}	

	/**
	 * Provide the attributes and the values searched for in the QuerySet.
	 * @param p_ElemType
	 * @param p_AttrSet
	 * @return SetList of UIDs.
	 */
	public SetList getNodesByAttr( String p_ElemType, Rslt2DArray p_AttrSet ) {
	
		boolean withCmp;
		
		int i, x;
		SetList UIDLst;
		SetList UIDLst2;
		
		String cmp;
		String qTail;

		String attrName = XMLDBDefs.c_TblNmAttrib + "." + XMLDBDefs.c_ColNmName;
		String attrVal = XMLDBDefs.c_TblNmAttrib + "." + XMLDBDefs.c_ColNmValue;
		String nodeUID = XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmUID;

		withCmp = p_AttrSet.getColSize() > 2;
		
		x = p_AttrSet.getRowNum();

		/* SELECT Node.UID 
			 FROM Node 
			 LEFT JOIN Attrib 
			 ON Node.UID = Attrib.UIDpNode 
			 WHERE Node.Name = 'TestCase' 
			 AND (Attrib.Name = 'Prio' AND Attrib.Value = '1') */

		String qMid = " FROM " + XMLDBDefs.c_TblNmNode;

		//LEFT JOIN  Attrib  
		qMid += " LEFT JOIN " + XMLDBDefs.c_TblNmAttrib;

		//ON Node.UID = Attrib.UIDpNod
		qMid += " ON " + nodeUID
			+ " = " + XMLDBDefs.c_TblNmAttrib + "." + XMLDBDefs.c_ColNmParentUID;

		// WHERE Node.Name = 'TestCase' 

		qMid += " WHERE " + XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmName
			+ " = '" + m_DB.escapeStr2( p_ElemType ) + "'";

		if( x > 0 ) {
			
			cmp = getCmpStmnt( attrVal, withCmp, p_AttrSet, 0 );
			
			// AND (Attrib.Name = 'Prio' AND Attrib.Value = '1')
			qTail = qMid 
				+ " AND (" + attrName + " = '" + p_AttrSet.getCell( 0, 0 ) + "'" 
				+ " AND " + cmp + ")";
		}
		else {
			
			qTail = qMid;		
		}

		UIDLst = m_DB.getKeys( nodeUID, qTail );
//UIDLst.addRsltAddInfo( qTail );			

		for( i = 1; (i < x) && UIDLst.isOK( true ); ++i ) {
			
			cmp = getCmpStmnt( attrVal, withCmp, p_AttrSet, i );

			// AND (Attrib.Name = 'Prio' AND Attrib.Value = '1')
			qTail = qMid 
				+ " AND (" + attrName + " = '" + p_AttrSet.getCell( 0, i ) + "'"
				+ " AND " + cmp + ")";

			UIDLst2 = m_DB.getKeys( nodeUID, qTail );

			if( UIDLst.checkRslt( UIDLst2, true ) ) {

				UIDLst.intersect( UIDLst2 );
			}
//UIDLst.addRsltAddInfo( qTail );				
		}

		if( UIDLst.notOK() ) {

			UIDLst.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodesByAttr(..)' failed." );
		}

		return UIDLst;
	}
	
	/**
	 * Provide the element type name, the attributes and the values searched and the list of 
	 * attributes and values to be excluded for in the QuerySet.
	 * @param p_ElemType
	 * @param p_AttrSet Provide rows with Name, compare value, compare operator and flag "isNumeric" as string.
	 * @param p_ExclAttrSet
	 * @return SetList of UIDs.
	 */
	public SetList getNodesByAttrExclByAttr( String p_ElemType, Rslt2DArray p_AttrSet, Rslt2DArray p_ExclAttrSet ) {
		
		boolean withCmp;
		
		int i, x;
		SetList UIDLst;
		SetList UIDLst2;
		String cmp = "";
		String qTail;

		
		String attrName = XMLDBDefs.c_TblNmAttrib + "." + XMLDBDefs.c_ColNmName;
		String attrVal = XMLDBDefs.c_TblNmAttrib + "." + XMLDBDefs.c_ColNmValue;
		String nodeUID = XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmUID;

		x = p_AttrSet.getRowNum();
		
		withCmp = p_AttrSet.getColSize() > 2;

		/* SELECT Node.UID 
			 FROM Node 
			 LEFT JOIN Attrib 
			 ON Node.UID = Attrib.UIDpNode 
			 WHERE Node.Name = 'TestCase' 
			 AND (Attrib.Name = 'Prio' AND Attrib.Value = '1') */

		String qMid = " FROM " + XMLDBDefs.c_TblNmNode;

		//LEFT JOIN  Attrib  
		qMid += " LEFT JOIN " + XMLDBDefs.c_TblNmAttrib;

		//ON Node.UID = Attrib.UIDpNod
		qMid += " ON " + nodeUID
			+ " = " + XMLDBDefs.c_TblNmAttrib + "." + XMLDBDefs.c_ColNmParentUID;

		// WHERE Node.Name = 'TestCase' 

		qMid += " WHERE " + XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmName
			+ " = '" + m_DB.escapeStr2( p_ElemType ) + "'";

//UIDLst = new SetList();

		if( x > 0 ) {
			
			cmp = getCmpStmnt( attrVal, withCmp, p_AttrSet, 0 );			
			
			// AND (Attrib.Name = 'Prio' AND Attrib.Value = '1')
			qTail = qMid 
				+ " AND (" + attrName + " = '" + p_AttrSet.getCell( 0, 0 ) + "'" 
				+ " AND " + cmp + ")";
		}
		else {
			
			qTail = qMid;
		}

		UIDLst = m_DB.getKeys( nodeUID, qTail );
//UIDLst.addRsltAddInfo( qTail );

//UIDLst.setInError( "-" );

		for( i = 1; (i < x) && UIDLst.isOK( true ); ++i ) {
			
			cmp = getCmpStmnt( attrVal, withCmp, p_AttrSet, i );
	
//UIDLst.addRsltAddInfo( cmp );
			
			// AND (Attrib.Name = 'Prio' AND Attrib.Value = '1')
			qTail = qMid 
				+ " AND (" + attrName + " = '" + p_AttrSet.getCell( 0, i ) + "'"
				+ " AND " + cmp + ")";
			
//UIDLst.addRsltAddInfo( qTail );

			UIDLst2 = m_DB.getKeys( nodeUID, qTail );

			if( UIDLst.checkRslt( UIDLst2, true ) ) {

				UIDLst.intersect( UIDLst2 );
			}
		}
		
		x = p_ExclAttrSet.getRowNum();
		
		for( i = 0; (i < x) && UIDLst.isOK( true ); ++i ) {

			// AND NOT (Attrib.Name = 'Prio' AND Attrib.Value = '1')
			qTail = qMid 
				+ " AND (" + attrName + " = '" + p_ExclAttrSet.getCell( 0, i ) + "'"
				+ " AND NOT ( " + attrVal + " = '" + m_DB.escapeStr2( p_ExclAttrSet.getCell( 1, i ) ) + "' ))";

			UIDLst2 = m_DB.getKeys( nodeUID, qTail );

			if( UIDLst.checkRslt( UIDLst2, true ) ) {

				UIDLst.intersect( UIDLst2 );
			}
		}

		if( UIDLst.notOK( true ) ) {

			UIDLst.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodesByAttrExclByAttr(..)' failed." );
		}

//UIDLst.addRsltAddInfo( "getNodesByAttrExclByAttr" );		
		return UIDLst;
	}

	/**
	 * @param p_ElemType
	 * @return RsltTable; Cols: UID, Name, Type
	 */
	public RsltTable getNodesByType( String p_ElemType ) {
	 
		RsltTable nodeTbl;

		String tblName = XMLDBDefs.c_TblNmNode;

		String[] cols = { XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmUID,
											XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmName,
											XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmNodeType };

		String appx = "WHERE ";

		// dump the dummy
		appx += " NOT (" +  XMLDBDefs.c_ColNmUID + " = '" + XMLDBDefs.c_DummyTopUID + "' )";

		if( p_ElemType.length() > 0 ) {

			appx += " AND " + XMLDBDefs.c_ColNmName + " = '" + p_ElemType + "'";		
		}

		nodeTbl = m_DB.select( tblName, cols, appx );

		if( nodeTbl.notOK() ) {

			nodeTbl.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodesByType(..)' failed." );
		}

		return nodeTbl;
	}

	/**
	 * Gets the value of a node.
	 * @param p_UID
	 * @return
	 */
	public RsltString getNodeVal( String p_UID ) { 

		RsltTable valTbl;
		RsltString val = new RsltString( true );

		String tblName = XMLDBDefs.c_TblNmNodeVal;

		String[] cols = { XMLDBDefs.c_ColNmValue };

		String appx = "WHERE " + XMLDBDefs.c_ColNmParentUID + " = '" + p_UID + "'";

		valTbl = m_DB.select( tblName, cols, appx );
		
		if( val.checkRslt( valTbl ) ) {

			val.setString( valTbl.getCell( 0, 0 ) );
		}
		else {

			val.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getNodeVal()' for '" + p_UID + "' failed." );
		}

		return val;
	}

	/**
	 * 
	 * @param p_UID
	 * @return
	 */
	public RsltString getParentID( RsltString p_UID ) {

		RsltTable parTbl;

		RsltString parUID;
		
		String pUID;

		String tblName = XMLDBDefs.c_TblNmNode;

		String[] cols = { XMLDBDefs.c_ColNmParentUID };

		String appx = "WHERE " + XMLDBDefs.c_ColNmUID + " = '" + p_UID.getString() + "'";

		parTbl = m_DB.select( tblName, cols, appx );

		parUID = new RsltString( true );
		
		if( parUID.checkRslt( parTbl ) ) {
			
			pUID = parTbl.getCell( 0, 0 );

			parUID.setString( pUID );
		}
					
		if( parUID.notOK() ) {

			parUID.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getParentID(..)' failed." );
		}

		return parUID;
	}

	/**
	 * @param p_UID
	 * @param p_SubNodeName
	 * @return
	 */
	public RsltTable getSubNodeVals( String p_UID, String p_SubNodeName ) {

		RsltTable valTbl;

		/* select nodeval.value from nodeval
				LEFT JOIN  node on nodeval.UIDpNode = node.UID
				WHERE node.Name = 'Title'
				AND node.UIDpNode = '00E07DC38C98-000039795DEE7E3A-00000000' ;*/

		String qHead = "SELECT ";

		String[] cols = { XMLDBDefs.c_TblNmNodeVal + "." + XMLDBDefs.c_ColNmValue };

		String qTail = " FROM " + XMLDBDefs.c_TblNmNodeVal
										+ " LEFT JOIN " + XMLDBDefs.c_TblNmNode 
										+ " ON " + XMLDBDefs.c_TblNmNodeVal + "." + XMLDBDefs.c_ColNmParentUID
										+ " = " + XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmUID 
										+ " WHERE " + XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmName
										+ " = '" + p_SubNodeName + "' AND "
										+ XMLDBDefs.c_TblNmNode + "." + XMLDBDefs.c_ColNmParentUID
										+ " = '" + p_UID + "'";

		valTbl = m_DB.freeSelect( qHead, cols, qTail );

		if( valTbl.notOK() ) {

			valTbl.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getSubNodeVals(..)' failed." );
		}

		return valTbl;
	}

	/**
	 * Gets the UID of the top node of the document, to which the node with the UID belongs to.
	 * @param p_UID
	 * @return The UID.
	 */
	public RsltString getTopNodeUID( RsltString p_UID ) {

		RsltString parUID;
		RsltString topUID;
		
		String pUID;

		parUID = getParentID( p_UID );
		
		if( parUID.notOK() ) {

			topUID = new RsltString();

			topUID.setRsltStat( parUID );
		}
		else {
			
			pUID = parUID.getString();

			if(  pUID.compareTo( XMLDBDefs.c_DummyTopUID  ) != 0 ) {

				topUID = getTopNodeUID( parUID );
			}
			else {

				topUID = p_UID;
			}
		}

		if( topUID.notOK() ) {

			topUID.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getTopNode([1])' failed." );
		}
		
		return topUID;
	}

	/**
	 * Gets the UID of the top node of the specified element type (if there is) of 
	 * the document, to which the node with the UID bleongs to. Walks up the tree 
	 * until it finds the requested element type.  
	 * @param p_tNodeElem
	 * @param p_UID
	 * @return The UID.
	 */
	public RsltString getTopNodeUID( RsltString p_tNodeElem, RsltString p_UID ) {

		RsltString parUID;
		RsltString topUID;
		RsltString nodeName;

		parUID = getParentID( p_UID );
		
		if( parUID.notOK() ) {

			topUID = new RsltString();

			topUID.setRsltStat( parUID );
		}
		else {

			if( parUID.getString() != XMLDBDefs.c_DummyTopUID ) {

				topUID = getTopNodeUID( p_tNodeElem, parUID );
			}
			else {

				nodeName = getNodeName( p_UID.getString() );

				if( nodeName.getString() != p_tNodeElem.getString() ) {

					topUID = new RsltString();
					topUID.setRsltStat( RsltStatVal.isEmpty, "" );
				}
				else {

					topUID = p_UID;
				}
			}
		}

		if( topUID.notOK() ) {

			topUID.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getTopNode([2])' failed." );
		}
		
		return topUID;
	}

	// -------------------------------------------
	protected RsltString getUID( Element p_Elem ) {
	 
		RsltString UID = new RsltString();
		
		String val;
		
		val = p_Elem.getAttribute( XMLDBDefs.c_ColNmUID );
		
		if( val.length() < 1 ) {

			UID = m_UIDServ.getUID();
		}
		else {
			
			UID.setString( val );
		}
				
		return UID;		
	}		
		
	// -------------------------------------------		
	protected RsltString getUID( Node p_Node ) {
		
		int attrColCnt = 0;
		
		NamedNodeMap attrCol;
		Node attrib;
		
		RsltString UID = new RsltString();
				
		boolean hasNoUID = true;

		try {
			
			attrCol = p_Node.getAttributes();
			
			if( null != attrCol ) {

				attrColCnt = attrCol.getLength();		
				
				if( attrColCnt > 0 ) {
					
					attrib = attrCol.getNamedItem( XMLDBDefs.c_ColNmUID );
	
					if( null != attrib ) {
	
						UID.setString( attrib.getNodeValue() );
	
						hasNoUID = false;
					}
				}
			}

			if( hasNoUID ) {

				UID = m_UIDServ.getUID();
			}
		}
		catch( Exception e ) {

			UID.setRsltByExcp( e );
		}
		
		if( UID.notOK() ) {

			UID.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getUID(..)' failed." );
		}	
			
		return UID;
	}

	/**
	 * Gets the first occurrence of a sub element of the given type.
	 * @param p_ElemType
	 * @param p_ParUID
	 * @return
	 */
	public RsltString getFirstUIDByParent( String p_ElemType, String p_ParUID ) {
		
		RsltString rUID = new RsltString();
		SetList uidLst;
		
		uidLst = getUIDLstByParent( p_ElemType, p_ParUID );
		
		if( uidLst.size() > 0 ) {
			
			rUID.setString( uidLst.getString( 0 ) );
		}
		
		return rUID;
	}
	
	/**
	 * Gets the list of nodes of the specified element type that belong to the parent UID.<br>
	 * If element type is empty (""), all subelement of all types are retrieved.
	 * @param p_ElemType
	 * @param p_ParUID
	 * @return list of UIDs
	 */
	public SetList getUIDLstByParent( String p_ElemType, String p_ParUID ) {

		SetList UIDLst;

		String col = XMLDBDefs.c_ColNmUID;

		String qTail = " FROM " + XMLDBDefs.c_TblNmNode;
		qTail += " WHERE " + XMLDBDefs.c_ColNmParentUID + " = '" + p_ParUID + "'";

		if( p_ElemType.length() < 1 ) {

			// dump the dummy
			qTail += " AND NOT (" + XMLDBDefs.c_ColNmUID + " = '" + XMLDBDefs.c_DummyTopUID + "' )";
		}
		else {

			qTail += " AND " + XMLDBDefs.c_ColNmName + " = '" + p_ElemType + "'";
		}

		UIDLst = m_DB.getKeys( col, qTail );

		/*
		if( m_DB.notOK() ) {

			UIDLst.setResStat( m_DB );
		}*/
		
		if( UIDLst.notOK( true ) ) {

			addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getUIDLstByParent(..)' failed." );
		}

		return UIDLst;
	}

	/**
	 * @param p_ParUID
	 * @return RsltTable; Cols: UID, Name, Type.
	 */
	public RsltTable getXNodeLstByParent( String p_ParUID ) {

		RsltTable nodeTbl;

		String tblName = XMLDBDefs.c_TblNmNode;

		String[] cols = { XMLDBDefs.c_ColNmUID, XMLDBDefs.c_ColNmName, XMLDBDefs.c_ColNmNodeType };

		String appx = "WHERE " + XMLDBDefs.c_ColNmParentUID + " = '" + p_ParUID + "'";

		nodeTbl = m_DB.select( tblName, cols, appx );

		if( nodeTbl.notOK() ) {

			nodeTbl.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.getXNodeLstByParent(..)' failed." );
		}

		return nodeTbl;
	}

	/**
	 * @param p_XMLdoc
	 * @return the UID; the given or a new one.
	 */
	public RsltString importXML( Document p_XMLdoc ){

		String dummyTopUID;
		int i, x;
		Node xNode;
		short nodeType;
		Rslt aRes = new Rslt( true );
		RsltString UID = new RsltString( true );
		String UIDstr;

		try {
			
			dummyTopUID = XMLDBDefs.c_DummyTopUID;
			
			Element root = p_XMLdoc.getDocumentElement();
			
			if( null  == root ) {
				
				UID.setInError( "No root node found." );
			}
			else {
				
				startTrans();
			}
			
			if( UID.isOK() ) {
				
				NodeList nodeLst = p_XMLdoc.getChildNodes();
				
				x = nodeLst.getLength();
				
				UID = getUID( root );
				
				UIDstr = UID.getString();
				
				// update by brute force
				RsltTable NodeTbl;
				NodeTbl = getNodeAsTable( UIDstr );
				
				if( NodeTbl.isNotEmpty() ) {
					
					aRes = deleteXML( UIDstr );
				}
				
				// we must create the data node first, cause all others will be dependant on it
				if( UID.checkRslt( aRes ) ) {
					
					UID.checkRslt( writeNodes( root, dummyTopUID, UIDstr ) );
					
					// this loop is supposed to write away the remaining meta-data of the xml
					for( i = 0; ((i < x) && isOK())  ; i++ ) {
						
						xNode = nodeLst.item(i);
						
						nodeType = xNode.getNodeType();
						
						if( Node.ELEMENT_NODE != nodeType ) {
							
							UID.checkRslt( writeAMetaNode( xNode, UIDstr ) );
						}
					}
				}
			}
		}
		catch( Exception excp ) {

			UID.setRsltByExcp( excp );
		}	

		if( UID.isOK() ) {

			commit();
		}
		else {

			UID.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.importXML(..)' failed." );
			rollBack();
		}

		return UID;
	}
	
	/**
	 * Service call to get a new UID.
	 * @return the UID.
	 */
	public RsltString newUID() {
		
		RsltString UID;
		
		UID = m_UIDServ.getUID();
		
		return UID;
	}

	/**
	 * The DB transaction is rolled back.
	 */
	protected void rollBack( ) { 

		m_DB.rollBack();

/* is already in error
		if( m_DB.notOK() ) {

			setResStat( m_DB );
			addResAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.startTrans()' failed." );
		}
*/		
	}

	/**
	 * A DB transaction is started.
	 */
	protected Rslt startTrans( ) { 

		Rslt res;
		
		res = m_DB.startTrans();

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.startTrans()' failed." );
		}
		
		return res;
	}

	// --------------------------------------------------------------------------
	protected Rslt writeAMetaNode( Node p_aNode, String p_UID ){

		int i, x;
		String aName;
		String aVal;
		short nodeType;
		Rslt aRes;
		Rslt res = new Rslt( true );
		RsltString nodeUID;
		String UIDstr;

		try {

			aName = p_aNode.getNodeName();
			nodeType = p_aNode.getNodeType();
			aVal = p_aNode.getNodeValue();
			//Console.WriteLine( aName );			

			switch( nodeType ) {

				case Node.COMMENT_NODE:

// currently we drop comments					
				break;

				case Node.DOCUMENT_TYPE_NODE:

					String bVal;
					
					DocumentType dType = p_aNode.getOwnerDocument().getDoctype();

					//bVal = p_aNode.OwnerDocument.DocumentType.PublicId;
					bVal = dType.getPublicId();

					if( null != bVal ) {

						aVal = "PUBLIC " + bVal;
					}

					bVal = dType.getSystemId();
					
					if( null != bVal ) {

						aVal += "SYSTEM " + bVal;
					}

					//aName = p_aNode.Value;???

					//we are at the top and this is the document node to the node identified by p_UID
					nodeUID = getUID( p_aNode );
					UIDstr = nodeUID.getString();

					writeNode2DB( aName, nodeType, p_UID, UIDstr );
					writeValNode2DB( UIDstr, aVal );
					
				break; 

				case Node.PROCESSING_INSTRUCTION_NODE:

					//we are at the top and this is a process instruction node to the node identified by p_UID
					//I do hope there are no process instructions node below the top node
					nodeUID = getUID( p_aNode );

					UIDstr = nodeUID.getString();

					writeNode2DB( aName, nodeType, p_UID, UIDstr );
					writeValNode2DB( UIDstr, aVal );
					break; 

/* Java does not know XmlDeclaration
				case Node.XmlDeclaration:

					//we are at the top and this is the xml declaration node to the node identified by p_UID
					nodeUID = getUID( p_aNode );
					UIDstr = nodeUID.getString();

					writeNode2DB( aName, nodeTypeInt, p_UID, UIDstr );
					writeValNode2DB( UIDstr, aVal );
					break; 
*/
				default:

					res.setInError( "'" + nodeType + "' is an unexpected type." );
					break; 
			}

			if( res.isOK() ) {

				NamedNodeMap AttrCol = p_aNode.getAttributes();

				if( AttrCol != null ) {

					x = AttrCol.getLength();

					for( i = 0; i < x; ++i ) {

						aName = AttrCol.item( i ).getNodeName();
						aVal = AttrCol.item( i ).getNodeValue();

						//Console.WriteLine( " " + aName + "=" + aVal	);
						aRes = writeAttrNode2DB( p_UID, aName, aVal );

						if( !res.checkRslt( aRes ) ) {

							i = x;
						}
					}
				}
			}
		}
		catch( Exception excp ) {

			res.setRsltByExcp( excp );
		}			

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeAMetaNode(..)' failed." );
		}
		
		return res;
	}

	// --------------------------------------------------------------------------
	protected RsltString writeANode( Node p_aNode, String p_parentUID ){

		//String aName;
		//short nodeType;
		
		Rslt aRes;
		
		RsltString UID;

		//aName = p_aNode.Name;
		//nodeType = p_aNode.getNodeType();
		//Console.WriteLine( aName );

		UID = getUID( p_aNode );

		if( UID.isOK() ) {
			
			String UIDstr = UID.getString();
			
			aRes = writeANode( p_aNode, p_parentUID, UIDstr );
			
			if( !UID.checkRslt( aRes ) ) {
				
				UID.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeANode([1])' failed." );
			}
		}

		return UID;
	}

	// --------------------------------------------------------------------------
	protected Rslt writeANode( Node p_aNode, String p_parentUID, String p_UID ){

		int i, x;
		
		Rslt aRes;
		Rslt res = new Rslt( true );
		
		String aName;
		String aVal;

		short nodeType;

		try {

			aName = p_aNode.getNodeName();
			nodeType = p_aNode.getNodeType();
			//aVal = p_aNode.Value;
			//Console.WriteLine( aName );			

			switch( nodeType ) {

				case Node.DOCUMENT_TYPE_NODE:
				case Node.PROCESSING_INSTRUCTION_NODE:
				//case Node.XmlDeclaration:

					res.setInError( "'" + nodeType + "' is an unexpected type." );
					break; 

				default:

					aRes = writeNode2DB( aName, nodeType, p_parentUID, p_UID );
				
					res.checkRslt( aRes );
					break; 
			}

			if( res.isOK() ) {

				aRes = delAttrs( p_UID );

				if( res.checkRslt( aRes ) ) {

					NamedNodeMap AttrCol = p_aNode.getAttributes();

					if( AttrCol != null ) {

						x = AttrCol.getLength();

						for( i = 0; i < x; ++i ) {

							aName = AttrCol.item( i ).getNodeName();

							if( XMLDBDefs.c_AttrNmUID != aName ) {

								aVal = AttrCol.item( i ).getNodeValue();

								//Console.WriteLine( " " + aName + "=" + aVal	);
								aRes = writeAttrNode2DB( p_UID, aName, aVal );

								if( !res.checkRslt( aRes ) ) {

									i = x;
								}
							}
						}
					}
				}
			}
		}
		catch( Exception excp ) {

			res.setRsltByExcp( excp );
		}	

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeANode([2])' failed." );
		}
		
		return res;
	}
	
	// --------------------------------------------------------------------------
	protected Rslt writeAttrNode2DB( String p_parentUID, String p_AttrName, String p_AttrVal ) {

		Rslt res;
		
		QuerySet qSet = new QuerySet( 3, XMLDBDefs.c_TblNmAttrib );

		qSet.addKeyCol( XMLDBDefs.c_ColNmParentUID, p_parentUID, false );
		
		qSet.addKeyCol( XMLDBDefs.c_ColNmName, p_AttrName, false );
		
		qSet.addQCol( XMLDBDefs.c_ColNmValue, p_AttrVal, true );
		
		res = m_DB.insertUpdate( qSet );

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeAttrNode2DB(..)' failed." );
		}
		
		return res;
	}

	// --------------------------------------------------------------------------
	public Rslt writeFileInfo( String p_UID, String p_XMLfile ) {

		Rslt res;
		//RsltString escDir;

		File xFile = new File( p_XMLfile );

		QuerySet qSet = new QuerySet( 3, XMLDBDefs.c_TblNmXMLFile );

		qSet.addKeyCol( XMLDBDefs.c_ColNmParentUID, p_UID, false );
		
		qSet.addQCol( XMLDBDefs.c_ColNmName, xFile.getName(), false );
		
		String rawDir =	xFile.getAbsolutePath();

		/*escDir = escapeString( rawDir );

		if( escDir.notOK() ) {

			res = escDir;
		}
		else {
		*/

		qSet.addQCol( XMLDBDefs.c_ColNmPath, rawDir, true );

		res = m_DB.insertUpdate( qSet );
		
		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeFileInfo(..)' failed." );
		}
		
		return res;
	}

	// --------------------------------------------------------------------------
	protected Rslt writeNode2DB( String p_NodeName, short p_NodeType, String p_ParentUID, String p_UID ) {

		Rslt res;
		
		QuerySet qSet = new QuerySet( 4, XMLDBDefs.c_TblNmNode );

		qSet.addKeyCol( XMLDBDefs.c_ColNmUID, p_UID, false );

		qSet.addKeyCol( XMLDBDefs.c_ColNmParentUID, p_ParentUID, false );

		qSet.addQCol( XMLDBDefs.c_ColNmName, p_NodeName, false );

		qSet.addQCol( XMLDBDefs.c_ColNmNodeType, String.valueOf( p_NodeType ), false );

		res = m_DB.insertUpdate( qSet );

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeNode2DB(..)' failed." );
		}
		
		return res;
	}

	// --------------------------------------------------------------------------
	protected RsltString writeNodes( Node p_aNode, String p_parentUID ){

		Rslt aRes;
		RsltString UID;
		String UIDstr;

		UID = getUID( p_aNode );

		UIDstr = UID.getString();

		aRes = writeANode( p_aNode, p_parentUID, UIDstr );

		if( UID.checkRslt( aRes ) ) {

			aRes = writeSubNodes( p_aNode, UIDstr );
			
			UID.checkRslt( aRes );
		}

		if( UID.notOK() ) {

			UID.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeNodes([2])' failed." );
		}

		return UID;
	}

	// --------------------------------------------------------------------------
	protected RsltString writeNodes( Node p_aNode, String p_parentUID, String p_UID ){

		Rslt aRes;
		RsltString resUID = new RsltString();

		resUID.setString( p_UID );
		
		aRes = writeANode( p_aNode, p_parentUID, p_UID );

		if( resUID.checkRslt( aRes ) ) {

			aRes = writeSubNodes( p_aNode, p_UID );
			
			resUID.checkRslt( aRes );
		}

		if( resUID.notOK() ) {

			addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeNodes([1])' failed." );
			resUID.setRsltStat( this );
		}

		return resUID;
	}

	// --------------------------------------------------------------------------
	protected Rslt writeSubNodes( Node p_aNode, String p_parentUID ){

		Rslt aRes;
		Rslt res = new Rslt( true );
		RsltString resUID = new RsltString();
		Node cNode;
		int i, x;

		NodeList cNodes = p_aNode.getChildNodes();

		XMLDBChildNodeLst existingChildLst = new XMLDBChildNodeLst( this, p_parentUID );

		if( res.checkRslt( existingChildLst ) ) {

			x = cNodes.getLength();
	
			for( i = 0; (i < x) && res.isOK(); ++i ) {

				cNode = cNodes.item( i );

				switch( cNode.getNodeType() ) {

					case Node.TEXT_NODE:

						aRes = writeTxtNode( cNode, p_parentUID );

						if( res.checkRslt( aRes ) ) {

							existingChildLst.removeFromLst( p_parentUID );
						}
					break; 

					default:

						resUID = writeNodes( cNode, p_parentUID );

						if( res.checkRslt( resUID ) ) {

							String UID = resUID.getString();

							existingChildLst.removeFromLst( UID );
						}
					break;
				}
			}
		}

		if( res.isOK() ) {

			existingChildLst.deleteRemNodes();
		}
		else {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeSubNodes(..)' failed." );
		}
		
		return res;
	}

	// --------------------------------------------------------------------------
	protected Rslt writeTxtNode( Node p_aTxtNode, String p_parentUID ){

		Rslt res;
		
		String aVal;

		aVal = p_aTxtNode.getNodeValue();

		//Console.WriteLine( " " + aVal	);

		res = writeValNode2DB( p_parentUID, aVal );
		
		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeTxtNode(..)' failed." );
		}
		
		return res;
	}

	// --------------------------------------------------------------------------
	protected Rslt writeValNode2DB( String p_parentUID, String p_aVal ) {

		Rslt res;
		
		//String escString;

		//escString = escapeString( p_aVal );

		QuerySet qSet = new QuerySet( 2, XMLDBDefs.c_TblNmNodeVal );

		qSet.addKeyCol( XMLDBDefs.c_ColNmParentUID, p_parentUID, false );

		qSet.addQCol( XMLDBDefs.c_ColNmValue, p_aVal, true );

		res = m_DB.insertUpdate( qSet );

		if( res.notOK() ) {

			res.addRsltAddInfo( "'com.qanavigation.XMLDB.XMLDBbasics.writeValNode2DB(..)' failed." );
		}
		
		return res;
	}
}
